//==================================================================//
//==================================================================//
//= Shining3D.cpp ==================================================//
//==================================================================//
//= Original coder: Trent Polack (ShiningKnight) ===================//
//= Code modifications by: Christian Ofenberg (AblazeSpace) ========//
//==================================================================//
//==================================================================//



//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- INCLUDES -------------------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//

#include "Shining3D.h"

//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- GLOBALS --------------------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//
SDL_Surface *surface;		//This is our SDL surface

LOG S3Dlog;			//The error log

//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- FUNCTION DECLARATIONS ------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//
int SHINING3D::ToggleFullScreen()
	{
	if (SDL_WM_ToggleFullScreen(surface))
		{
		fullscreen=!fullscreen;
		if(fullscreen==FULLSCREEN_ON)		//Did the user want a fullscreen window?
			SDL_ShowCursor(SDL_DISABLE);	//Hide the mouse pointer
		else
			SDL_ShowCursor(SDL_ENABLE);		//Show the mouse pointer
		return 1;
		}

	return 0;
	}

//------------------------------------------------------------------//
//- bool SHINING3D:: OpenGL_Init(int, int, int, int) ---------------//
//------------------------------------------------------------------//
//- Description: This function takes care of all the necessary	   -//
//-              OpenGL initiation things.  You can provide the    -//
//-              dimensions (width, height, and bits per pixel),   -//
//-              and you can also make it a window, fullscreen or  -//
//-              give the user a choice, using these flags:	       -//
//-                  - FULLSCREEN (Just start with a fullscreen)   -//
//-                  - WINDOW (Just start as a window)             -//
//-                  - GIVE_CHOICE ("Would you like to run in fullscreen mode?") -//
//------------------------------------------------------------------//
bool SHINING3D::OpenGLInit(int width, int height, int bpp, int screenflag)
	{
	/* this holds some info about our display */
	const SDL_VideoInfo *videoInfo;
	
	SCREEN_WIDTH =width;			//Set the class's constants to the
	SCREEN_HEIGHT=height;			//parameter's provided by the 
	SCREEN_BPP   =bpp;				//user.

	S3Dlog.Output("Window initiation:");

	/* initialize SDL */
	if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
		{
		S3Dlog.Output(" Video initialization failed: %s", SDL_GetError());
		SDL_Quit();
		return false;
		}
	S3Dlog.Output(" Video initialization ... OK");
	
	/* Fetch the video info */
	videoInfo = SDL_GetVideoInfo( );

	if ( !videoInfo )
		{
		S3Dlog.Output(" Video query failed: %s", SDL_GetError());
		SDL_Quit();
		return false;
		}
	S3Dlog.Output(" Video query ... OK");

  	/* the flags to pass to SDL_SetVideoMode */
	videoFlags  = SDL_OPENGL;          /* Enable OpenGL in SDL */
	videoFlags |= SDL_HWPALETTE;       /* Store the palette in hardware */
	videoFlags |= SDL_RESIZABLE;       /* Enable window resizing */

	/* This checks to see if surfaces can be stored in memory */
	if ( videoInfo->hw_available )
		videoFlags |= SDL_HWSURFACE;
	else
		videoFlags |= SDL_SWSURFACE;

	/* This checks if hardware blits can be done */
	if ( videoInfo->blit_hw )
		videoFlags |= SDL_HWACCEL;

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,        1);	// tell SDL that the GL drawing is going to be double buffered
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, SCREEN_BPP);	// size of depth buffer
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE,        0);	// we aren't going to use the stencil buffer
	SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE,      0);	// this and the next three lines set the bits allocated per pixel -
	SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE,    0);	// - for the accumulation buffer to 0
	SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE,     0);
	SDL_GL_SetAttribute(SDL_GL_ACCUM_ALPHA_SIZE,    0);

	fullscreen = FULLSCREEN_OFF;
	if(screenflag==GIVE_CHOICE)
		{
		//Ask the user which screen mode they prefer
		char cRis;
		cout << "Would you like to run in fullscreen mode (Y/N)? ";
		cin >> cRis;
		if (cRis == 'Y' || cRis == 'y')
			screenflag=FULLSCREEN;			//Fullscreen mode
		else
			screenflag=WINDOW;				//Windowed mode
		}

	/* get a SDL surface */
	surface = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, videoFlags );

	/* Verify there is a surface */
	if ( !surface )
		{
		S3Dlog.Output(" Video mode set failed: %s", SDL_GetError());
		SDL_Quit();
		return false;
		}
	S3Dlog.Output(" Video mode set ... OK");

	SDL_WM_SetCaption(TITLE, TITLE);	// set the window caption (first argument) and icon caption (2nd arg)

	if(screenflag==FULLSCREEN)				//Attempt fullscreen mode?
		{
		//Try to set selected mode and get results
		if (!ToggleFullScreen())
			{
			//If the mode fails, offer 2 options. Quit or use windowed mode.
			char cRis;
			cout << "The Requested Fullscreen Mode Is Not Supported By Your Video Card." << endl <<	"Use Windowed Mode Instead (Y/N)? ";
			cin >> cRis;
			if (cRis == 'Y' || cRis == 'y')
				{
				S3Dlog.Output("Fullscreen mode not supported.");
				}
			else
				{
				//Display a message letting the user know the program is closing.
				cout << "Program Will Now Close." << endl;
				SDL_Quit();
				return false;
				}
			}
		}

	glEnable(GL_TEXTURE_2D);							//Enable two dimensional texture mapping
	glShadeModel(GL_SMOOTH);							//Enable smooth shading (so you can't see the individual polygons of a primitive, best shown when drawing a sphere)
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);				//Completely black background

	glClearDepth(1.0);									//Depth buffer setup
	glEnable(GL_DEPTH_TEST);							//Enable depth testing
	glDepthFunc(GL_LEQUAL);								//The type of depth testing to do (LEQUAL==less than or equal to)

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	//The nicest perspective look

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);					//Full brightness.
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);					//Set the blending function for transparency

	ResizeGLScene(width, height);						//Set up the OpenGL perspective view

	return true;
	}

//------------------------------------------------------------------//
//- bool SHINING3D:: OpenGL_Shutdown(void) -------------------------//
//------------------------------------------------------------------//
//- Description: This function takes care of all the necessary     -//
//-              OpenGL shutdown precautions.					   -//
//------------------------------------------------------------------//
void SHINING3D::OpenGLShutdown(void)
	{
	if(fullscreen==FULLSCREEN_ON)			//Are we in fullscreen mode?
		{
		ToggleFullScreen();					//If so switch back to the desktop and show mouse pointer
		}

	S3Dlog.Output("Shutdown:");	

	SDL_Quit();
	}

//------------------------------------------------------------------//
//- GLvoid SHINING3D:: Resize_GLScene(GLsizei, GLsizei) ------------//
//------------------------------------------------------------------//
//- Description: This function resizes the OpenGL SCENE (not the   -//
//-              window) so that the same proportions in the scene -//
//-              are the same.									   -//
//------------------------------------------------------------------//
GLvoid SHINING3D::ResizeGLScene(GLsizei width, GLsizei height)
	{
	if(height==0)										//Prevent a divide by zero (bad)
		height=1;										//Making height equal one

	glViewport(0,0,width,height);						//Reset the current viewport

	glMatrixMode(GL_PROJECTION);						//Select the projection matrix
	glLoadIdentity();									//Reset the projection matrix


	//Calculate the aspect ratio of the window
	gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,1.0f,10000.0f);

	glMatrixMode(GL_MODELVIEW);							//Select the modelview matrix
	glLoadIdentity();									//Reset The modelview matrix
	}

//------------------------------------------------------------------//
//- void SHINING3D:: HandleMessages(void) --------------------------//
//------------------------------------------------------------------//
//- Description: This function handles all of the necessary        -//
//-              messages that your window will recieve. This      -//
//-              function is more of a 'space-saver' for to reduce -//
//-              the amount of stuff you have to look through in   -//
//-              your console.									   -//
//------------------------------------------------------------------//
bool SHINING3D::HandleMessages(void)
{
	/* used to collect events */
	SDL_Event event;

	/* handle the events in the queue */
	while ( SDL_PollEvent( &event ) )
	{
		switch( event.type )
		{
			case SDL_VIDEORESIZE:
			{
				/* handle resize event */
				surface = SDL_SetVideoMode( event.resize.w, event.resize.h, SCREEN_BPP, videoFlags );
				if ( !surface )
				{
					S3Dlog.Output(" Could not get a surface after resize: %s", SDL_GetError());
					return false;
				}
				ResizeGLScene(event.resize.w, event.resize.h);
				break;
			}
			case SDL_QUIT:
			{
				/* handle quit requests */
				return false;
				break;
			}
			default:
				break;
		}
	}
	return true;
}

//------------------------------------------------------------------//
//- GLvoid SHINING3D::Font_Init(GLvoid) ----------------------------//
//------------------------------------------------------------------//
//- Description: This function initiates the font engine. Just put -//
//-              a call to this function in the initiation part of -//
//-              your program.									   -//
//------------------------------------------------------------------//
bool SHINING3D::FontInit(GLvoid)	
	{
	float cx=0;
	float cy=0;
	int loop;

	if (!font.LoadTGA("data/font.tga", GL_LINEAR, GL_LINEAR))
		return false;

	base=glGenLists(95);						//Creating a display list that can hold 95
	glBindTexture(GL_TEXTURE_2D, font.ID);		//Bind our font texture
	for(loop=0; loop<95; loop++)				//Loop through the display list
		{
		cx=float(loop%16)/16.0f;				//X position of current character
		cy=float(loop/16)/8.0f;					//Y position of current character

		glNewList(base+loop,GL_COMPILE);		//Start building a list
			glBegin(GL_TRIANGLE_STRIP);			//Use a quad for each character
				//Top right (tex/vertex) coordinates
				glTexCoord2f( (cx+0.0830f), (1.0f-cy));		
				glVertex2i(16,16);
				//Top left (tex/vertex) coordinates
				glTexCoord2f( (cx),			(1.0f-cy));	
				glVertex2i(0,16);
				//Bottom right (tex/vertex) coordinates
				glTexCoord2f( (cx+0.0830f), (1.0f-cy-0.120f)); 
				glVertex2i(16,0);
				//Bottom left (tex/vertex) coordinates
				glTexCoord2f( (cx),			(1.0f-cy-0.120f)); 
				glVertex2i(0,0);
			glEnd();							//Done Building Our Quad (Character)
			glTranslated(10,0,0);				//Move To The Right Of The Character
		glEndList();							//Done Building The Display List
		}															// Loop Until All 256 Are Built

	return true;
	}

//------------------------------------------------------------------//
//- GLvoid SHINING3D:: Font_Shutdown(GLvoid) -----------------------//
//------------------------------------------------------------------//
//- Description: This function shuts down the font engine.  Put a  -//
//-              call to this in the shutdown part of your program.-//
//------------------------------------------------------------------//
GLvoid SHINING3D::FontShutdown(GLvoid)		
	{
	glDeleteLists(base, 95);
	}

//------------------------------------------------------------------//
//- GLvoid SHINING3D::glPrint(GLint, GLint, const char*, ...) ------//
//------------------------------------------------------------------//
//- Description: This function prints a string of text onto the    -//
//-				 the screen.  You *must* be in ortho view to use   -//
//-              this function, but thats not a bad thing, it makes-//
//-              positioning the text easier.  Also, this function -//
//-              functions (hehehe) almost exactly like printf(...)-//
//-              in dos.										   -//
//------------------------------------------------------------------//
GLvoid SHINING3D::glPrint(GLint x, GLint y, const char *string, ...)
	{
	char		text[256];								//Holds Our String
	va_list		va;										//Pointer To List Of Arguments

	if(string==NULL)									//If There's No Text
		return;											//Do Nothing

	va_start(va, string);								//Parses The String For Variables
	    vsprintf(text, string, va);						//And Converts Symbols To Actual Numbers

	va_end(va);											//Results Are Stored In Text

	glBindTexture(GL_TEXTURE_2D, font.ID);				//Select Our Font Texture
	glPushMatrix();										//Store The Modelview Matrix
	glLoadIdentity();									//Reset The Modelview Matrix
	glTranslated(x,y,0);								//Position The Text (0,0 - Bottom Left)
	glListBase(base-32);								//Choose The Font Set
	glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);	//Draws The Display List Text
	glPopMatrix();										//Restore The Old Projection Matrix
	}

//------------------------------------------------------------------//
//- GLvoid LOG::Init() ---------------------------------------------//
//------------------------------------------------------------------//
//- Description: Don't even bother using this function at all, it  -//
//-              initiates the output log, but I put this function -//
//-              in the log's constructor, so it's automatically   -//
//-              called.										   -//
//------------------------------------------------------------------//
bool LOG::Init(void)	
	{
	FILE* logfile;
	//Clear the log contents
	if((logfile=fopen("Shining3D_Log", "wb"))==NULL)
		return false;

	//Close the file, and return a success!
	fclose(logfile);
	return true;
	}

//------------------------------------------------------------------//
//- GLvoid LOG::Output() -------------------------------------------//
//------------------------------------------------------------------//
//- Description: This function outputs text the error log, the     -//
//-              function functions (I won't even say "hehehe" this-//
//-              time) exactly like printf(...).				   -//
//------------------------------------------------------------------//
bool LOG::Output(char* text, ...)
	{
	FILE* logfile;
	va_list arg_list;

	//Initialize variable argument list
	va_start(arg_list, text);

	//Open the log file for append
	if((logfile = fopen("Shining3D_Log", "a+"))==NULL)
		return false;

	//Write the text and a newline
	vfprintf(logfile, text, arg_list);
	putc('\n', logfile);

	//Close the file
	fclose(logfile);
	va_end(arg_list);

	return true;
	}
