//==================================================================//
//==================================================================//
//= Shining3D.h ====================================================//
//==================================================================//
//= Original coder: Trent Polack (ShiningKnight) ===================//
//= Code modifications by: Christian Ofenberg (AblazeSpace) ========//
//==================================================================//
//==================================================================//

#ifndef Shining3D_H
#define Shining3D_H

//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- INCLUDES -------------------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//

#include <iostream>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glu.h>			//The OpenGL Glu header file
#include <SDL/SDL.h>
#include <SDL/SDL_syswm.h>
#include "Math.h"

//------------------------------------------------------------------//

//------------------------------------------------------------------//
//- CONSTANTS ------------------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//
#define TITLE "Stars 3D"
#define CLASSNAME "3dShooter"

/* Window type constants (used with the fullscreen flag) */
#define WINDOW		0
#define FULLSCREEN  1
#define GIVE_CHOICE 2

/* Fullscreen flags */
#define FULLSCREEN_ON  1
#define FULLSCREEN_OFF 0

#define PI		   3.1415926535897932384626433832795f
#define PIOVER180  0.0174532925199432957692369076848861f
#define PIUNDER180 57.2957795130823208767981548141052f
#define	VERTEX3D VECTOR3D

#define BITMAP_ID 0x4D42

using namespace std;

typedef unsigned long DWORD;
typedef float         FLOAT;

/* MD2 constants */
#define MAX_TRIANGLES 4096
#define MAX_VERTICES  2048
#define MAX_TEXCOORDS 2048
#define MAX_FRAMES    512

#define CHUNK1 0
#define CHUNK2 1
#define CHUNK3 2
#define CHUNK4 3

/* MD2 state constants */
#define ANIMATE_CURRENT -100

#define IDLE1				0
#define RUN					1
#define SHOT_STAND			2
#define SHOT_SHOULDER		3
#define JUMP				4
#define IDLE2				5
#define SHOT_FALLDOWN		6
#define IDLE3				7
#define IDLE4				8
#define CROUCH				9
#define CROUCH_CRAWL		10
#define CROUCH_IDLE			11
#define CROUCH_DEATH		12
#define DEATH_FALLBACK		13
#define DEATH_FALLFORWARD	14
#define DEATH_FALLBACKSLOW	15

#define IDLE1_START					0
#define IDLE1_END					39
#define RUN_START					40
#define RUN_END						45
#define SHOT_STAND_START			46
#define SHOT_STAND_END				60
#define SHOT_SHOULDER_START			61
#define SHOT_SHOULDER_END			66
#define JUMP_START					67
#define JUMP_END					73
#define IDLE2_START					74
#define IDLE2_END					95
#define SHOT_FALLDOWN_START			96
#define SHOT_FALLDOWN_END			112
#define IDLE3_START					113
#define IDLE3_END					122
#define IDLE4_START					123
#define IDLE4_END					135
#define CROUCH_START				136
#define CROUCH_END					154
#define CROUCH_CRAWL_START			155
#define CROUCH_CRAWL_END			161
#define CROUCH_IDLE_START			162
#define CROUCH_IDLE_END				169
#define CROUCH_DEATH_START			170
#define CROUCH_DEATH_END			177
#define DEATH_FALLBACK_START		178
#define DEATH_FALLBACK_END			185
#define DEATH_FALLFORWARD_START		186
#define DEATH_FALLFORWARD_END		190
#define DEATH_FALLBACKSLOW_START	191
#define DEATH_FALLBACKSLOW_END		198

/* MS3D constants */
#define MS3D_TEXTURE_OFF false
#define MS3D_TEXTURE_ON  true

//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- STRUCTURES -----------------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//	

//------------------------------------------------------------------//
//- MD2 STRUCTURES -------------------------------------------------//
//------------------------------------------------------------------//

typedef struct MD2_HEADER_TYP
	{
	int magic;				//The 'magic' number (always 844121161)
	int version;			//The version number of the file (always 8)
	int skinWidth;			//The width of the model's skin
	int skinHeight;			//The height of the model's skin
	int frameSize;			//The size of each frame (in BYTEs)
	int numSkins;			//The model's number of skins
	int numVertices;		//The model's number of vertices
	int numTexcoords;		//The model's number of texture coordinates (most likely, its not the same number as the vertices)
	int numTriangles;		//The number of triangles that make up the model
	int numGlCommands;		//The number of DWORDs (4 bytes) in the GLcommand list (which decide to render the model with tri strips, or fans)
	int numFrames;			//The number of frames (of animation) that the model has
	int offsetSkins;		//Offset in the file of the model, to where the skin information is
	int offsetTexcoords;	//Offset in the file of the model, to where the texture coordinate information is
	int offsetTriangles;	//Offset in the file of the model, to where the traingle information is
	int offsetFrames;		//Offset in the file of the model, to where the first frame information is given
	int offsetGlCommands;	//Offset in the file of the model, to where the GLcommand information is
	int offsetEnd;			//Offset in the file of the model, to where the end of it is
	} MD2_HEADER, *MD2_HEADER_PTR;

typedef struct MD2_VERTEX_TYP
	{
	unsigned char vertex[3];		//Scaled version of the model's 'real' vertex coordinate
	unsigned char lightNormalIndex;	//An index into the table of normals, kept by Quake 2
	} MD2_VERTEX, *MD2_VERTEX_PTR;

typedef struct MD2_FRAME_TYP
	{
	float	   scale[3];		//The scale used by the model's 'fake' vertex structure
	float	   translate[3];	//The translation used by the model's 'fake' vertex structure
	char	   name[16];		//The name for the frame
	MD2_VERTEX vertices[1];		//An array of MD2_VERTEX structures
	} MD2_FRAME, *MD2_FRAME_PTR;

typedef struct MD2_MODELVERTEX_TYP
	{
	float x,y,z;				//The (x,y,z) location of the vertex
	float u,v;					//The (u,v) location of the texture coordinate
	} MD2_MODELVERTEX, *MD2_MODELVERTEX_PTR;

//------------------------------------------------------------------//
//- MS3D STRUCTURES ------------------------------------------------//
//------------------------------------------------------------------//
// File header
typedef struct MS3D_HEADER_TYP
	{
    char id[10];				//The ID number of the model (always "MS3D000000")
	int  version;				//The version number of the model
	} MS3D_HEADER, *MS3D_HEADER_PTR;

typedef struct MS3D_VERTEX_TYP
	{
	unsigned char flags;

	unsigned char refCount;
	char boneID;				//The bone ID (used for skeletal animation)
	float vertex[3];
	} MS3D_VERTEX, *MS3D_VERTEX_PTR;

typedef struct MS3D_TRIANGLE_TYP
	{
	unsigned short flags;
	unsigned short vertexIndices[3];
	float vertexNormals[3][3];
	float u[3];
	float v[3];
	unsigned char smoothingGroup;
	unsigned char groupIndex;
	} MS3D_TRIANGLE, *MS3D_TRIANGLE_PTR;

typedef struct MS3D_GROUP_TYP
	{
    unsigned char	flags;
    char			name[32];
    unsigned short	numTriangles;
	unsigned short*	triangleIndices;
	char			materialIndex;
	} MS3D_GROUP, *MS3D_GROUP_PTR;

//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- CLASSES --------------------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//
class TEXTURE;
class MS3D;
class MD2;
class SHINING3D;
class SHININGINPUT;
class LOG;

//------------------------------------------------------------------//
//- class TEXTURE --------------------------------------------------//
//------------------------------------------------------------------//
//- The main texture class, has loading support for truevision	   -//
//- targas (.tga), and the Microsoft Bitmap (.bmp) format.		   -//
//------------------------------------------------------------------//
class TEXTURE
	{
	public:
		GLubyte* data;			//Image data (up to 32 bits)
		unsigned int bpp;		//Image color depth in bits per pixel
		unsigned int width;		//Image width
		unsigned int height;	//Image height
		unsigned int ID;		//Texture ID used to select a texture

   	bool LoadTGA(char* filename, GLfloat minFilter, GLfloat maxFilter);
   	bool LoadBMP(char* filename, GLfloat minFilter, GLfloat maxFilter);

   	TEXTURE();
   	~TEXTURE();
	};

//------------------------------------------------------------------//
//- class MS3D -----------------------------------------------------//
//------------------------------------------------------------------//
//- The native model format for Milkshape (3D modeler), this is a  -//
//- great format to use because Milkshape has support for some of  -//
//- most popular model formats around.  In short: use it.	       -//
//------------------------------------------------------------------//
class MS3D
	{
	public:
		unsigned short numVertices;
		MS3D_VERTEX*   vertices;
		unsigned short numTriangles;
		MS3D_TRIANGLE* triangles;
		unsigned short numGroups;
		MS3D_GROUP*	   groups;

	bool Load(char* filename);
	void Render(void);

	MS3D()
		{	}

	~MS3D()
		{	
		if(vertices)
			delete vertices;		
		if(triangles)
			delete triangles;
		if(groups)
			delete groups;
		}
	};

//------------------------------------------------------------------//
//- class MD2 ------------------------------------------------------//
//------------------------------------------------------------------//
//- A class for the Quake 2 model mesh (includes key frames for    -//
//- animation).													   -//
//------------------------------------------------------------------//
class MD2
	{
	private:	
		int numGlCommands;
		long* glCommands;

		int numTriangles;

	public:
		int stateStart;

		int stateEnd;

		int frameSize;
		int numFrames;
		char* frames;

		int currentFrame;
		int nextFrame;
		int endFrame;
		float interpolation;

	bool Load(char* filename);
	void Render(int numFrame);
	void Animate(int startFrame, int EndFrame, float Interpolation);
	void SetState(int state);

	MD2() : numGlCommands(0), stateStart(IDLE1_START), stateEnd(IDLE1_END),
			frameSize(0), numFrames(0), 
			currentFrame(IDLE1_START), nextFrame(currentFrame+1),
			endFrame(IDLE1_END), interpolation(0.0f)
		{	}

	~MD2()
		{	
		if(glCommands)
			delete [] glCommands;
		if(frames)
			delete [] frames;
		}
	};

//------------------------------------------------------------------//
//- class SHINING3D ------------------------------------------------//
//------------------------------------------------------------------//
//- The main graphics/window class.  Has functions for creating a  -//
//- window, and various other stuff.							   -//
//------------------------------------------------------------------//
class SHINING3D
	{
	private:
		/* Flags to pass to SDL_SetVideoMode */
		int videoFlags;
		TEXTURE font; //Font texture

		unsigned int base;

	public: 
		int SCREEN_WIDTH;
		int SCREEN_HEIGHT;
		int SCREEN_BPP;
		bool fullscreen;

	bool HandleMessages(void);
	void OpenGLShutdown(void);
	int  ToggleFullScreen(void);
	bool OpenGLInit(int width, int height, int bpp, int screenflag);
	inline void BufferSwap(void)
		{	SDL_GL_SwapBuffers();	}

	/* Font functions (PUBLIC) */
	GLvoid glPrint(GLint x, GLint y, const char *string, ...);	
	bool   FontInit(GLvoid);
	GLvoid FontShutdown(GLvoid);

	GLvoid ResizeGLScene(GLsizei width, GLsizei height);
	
	SHINING3D()	{}
	~SHINING3D() {}
	};

//------------------------------------------------------------------//
//- class SHININGINPUT ---------------------------------------------//
//------------------------------------------------------------------//
//- The input class, handles all input under SDL.  Use the -//
//- KEY_DOWN macro to see if a key is down or not.				   -//
//------------------------------------------------------------------//
class SHININGINPUT
{
	private:
		Uint8 *keyBuffer;	//The keyboard's key buffer

	public:
	SHININGINPUT();
	~SHININGINPUT();
	void SDLInputUpdate(void);
	/** Tests to see if a key is currently pressed down or not.
		Uses the SDL SDLK_key constants. */
	bool KeyDown(SDLKey key);
};

//------------------------------------------------------------------//
//- class LOG ------------------------------------------------------//
//------------------------------------------------------------------//
//- Creates a log text file, to keep track of what went successful,-//
//- what went wrong, and other information.  Defined in Log.cpp.   -//
//------------------------------------------------------------------//
class LOG
	{
	private:

	bool Init(void);

	public:

	bool Output(char* text, ...);

	LOG()
		{ Init(); }

	~LOG()
		{}
	};

//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- GLOBALS --------------------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//

extern SDL_Surface *surface;	//This is our SDL surface

extern SHININGINPUT sinput;		//To handle the input

extern LOG S3Dlog;

//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- MACROS ---------------------------------------------------------//
//------------------------------------------------------------------//
//------------------------------------------------------------------//
//Andr's macros. Thanks a lot.  

#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
#define MAX(a, b)  (((a) > (b)) ? (b) : (a))

#define RANDOM_FLOAT (((float)rand()-(float)rand())/RAND_MAX)
#define CHECK_RANGE(x,min,max) ((x= (x<min  ? min : x<max ? x : max)))
#define DEG_TO_RAD(angle)  ((angle)*PIOVER180)
#define RAD_TO_DEG(radians) ((radians)*PIUNDER180)
#define SQUARE(number) (number*number)

//------------------------------------------------------------------//
//------------------------------------------------------------------//
//- FUNCTION DECLARATIONS ------------------------------------------//
//------------------------------------------------------------------//

//------------------------------------------------------------------//

#endif	//Shining3D.h
